#   Copyright 2022 Xieyang Xu
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from typing import Dict, List, Set
import ipaddress
from pybatfish.datamodel.route import NextHop, NextHopInterface


from .utils import get_prefix_len, convert_prefix
#from .network import InterfaceConfig

class RibRule:
    def __init__(self, prefix: str, nexthop: NextHop) -> None:
        self.network: str = prefix
        self.prefix: ipaddress.IPv4Network = convert_prefix(prefix)
        self.nexthop: NextHop = nexthop

    def __repr__(self) -> str:
        return f"{self.network} -> {self.nexthop}"

    def __eq__(self, _o) -> str:
        return isinstance(_o, RibRule) and self.network == _o.network and self.nexthop == _o.nexthop

    def __hash__(self) -> int:
        return hash((self.network, str(self.nexthop)))

class SortedRib:
    def __init__(self) -> None:
        self.rules: List[RibRule] = []
        self.sorted = False
        self.forward_cache: Dict[str, Set[str]] = {}
        self.matched_rules_cache: Dict[str, Set[RibRule]] = {}

    def add_rule(self, prefix: str, nexthop: NextHop):
        self.rules.append(RibRule(prefix, nexthop))
        self.sorted = False

    def sort(self):
        """sort fib rules w.r.t. prefixlen (in descending order)"""
        self.rules.sort(key=get_prefix_len, reverse=True)
        self.sorted = True

    def lpm(self, ip: str) -> List[RibRule]:
        """find LPM rules. Multipath routing allowed."""
        ipa = ipaddress.ip_address(ip)
        if not self.sorted:
            self.sort()

        res = []
        longest_prefixlen = 0
        for rule in self.rules:
            if ipa in rule.prefix:
                if rule.prefix.prefixlen >= longest_prefixlen:
                    res.append(rule)
                    longest_prefixlen = rule.prefix.prefixlen
                else:
                    break
        return res

    def forward(self, ip: str) -> Set[str]:
        """find next hop interfaces. Multipath forwarding allowed."""
        if ip in self.forward_cache:
            return self.forward_cache[ip]

        res = set()
        dirty_ips = [ip]
        while dirty_ips:
            working_ip = dirty_ips.pop(0)
            for rule in self.lpm(working_ip):
                if rule.nexthop.type == "ip":
                    dirty_ips.append(rule.nexthop.ip)
                elif rule.nexthop.type == "interface":
                    res.add(rule.nexthop.interface)
        self.forward_cache[ip] = res
        return res

    def matched_rules(self, ip: str) -> Set[RibRule]:
        """find rules used to forward an ip"""
        if ip in self.matched_rules_cache:
            return self.matched_rules_cache[ip]

        res = set()
        dirty_ips = [ip]
        while dirty_ips:
            working_ip = dirty_ips.pop(0)
            for rule in self.lpm(working_ip):
                if rule.nexthop.type == "ip":
                    dirty_ips.append(rule.nexthop.ip)
                    res.add(rule)
                elif rule.nexthop.type == "interface":
                    res.add(rule)
        self.matched_rules_cache[ip] = res
        return res



#   Copyright 2022 Xieyang Xu
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import logging
import re
from parse import parse

from ..datamodel.dnode import *
from ..datamodel.network import *

def find_peer_session(network: Network, input_session: BgpSessionStatus) -> Optional[BgpSessionStatus]:
    peer_name = input_session.peer
    peer_device = network.devices[peer_name]
    peer_session = peer_device.find_bgp_session_with_as_ip(input_session.local_as, input_session.local_ip)
    return peer_session


def retrieve_leaf_vlans(bf, regex="/leaf/"):
    """Get a vlan prefix for each leaf in the network.
    @arg regex: a filter for leaf name

    >>> retrieve_leaf_vlans()
    ['leaf01', 'leaf02'], {'leaf01': '10.100.1.0/24', 'leaf02': '10.100.2.0/24'}
    """
    leaves = bf.q.nodeProperties(nodes=regex).answer().frame()
    leaves = leaves["Node"].to_list()
    vlans = {}

    for leaf in leaves:
        vlan_records = bf.q.switchedVlanProperties(nodes=leaf).answer().frame().to_records()
        if len(vlan_records) > 0:
            for interface in vlan_records[0].Interfaces:
                info_records = bf.q.interfaceProperties(nodes=leaf, interfaces=interface.interface).answer().frame().to_records()
                for rec in info_records:
                    if len(rec.All_Prefixes) > 0:
                        prefix = rec.All_Prefixes[0]
                        if not leaf in vlans:
                            vlans[leaf] = prefix

    return leaves, vlans

def retrieve_leaves(bf, regex="/leaf/"):
    """Get a vlan prefix for each leaf in the network.
    @arg regex: a filter for leaf name

    >>> retrieve_leaves()
    ['leaf01', 'leaf02'], {'leaf01': '10.100.1.0/24', 'leaf02': '10.100.2.0/24'}
    """
    leaves = bf.q.nodeProperties(nodes=regex).answer().frame()
    leaves = leaves["Node"].to_list()
    return leaves

def find_composed_peer_policy(policies: List[str], ip: str, direction: str) -> Optional[str]:
    matched = []
    for policy in policies:
        if re.search(ip, policy) and re.search(direction, policy):
            matched.append(policy)
    if len(matched) > 1:
        logging.getLogger(__name__).warning(f"WARNING: expect unique policy for {ip} {direction}, actual matches: {matched}")
    return matched[0] if len(matched) > 0 else None

def extract_bgp_neighbor_ip_vrf(name: str) -> Tuple[str, str]:
    return parse("{} (VRF {})", name)

def fraction_repr(covered: int, all: int) -> str:
    return f"{covered}/{all} ({'{:.2%}'.format(covered/all) if all != 0 else '0.00%'})"


#   Copyright 2022 Xieyang Xu
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from __future__ import annotations
import logging # type hint in the enclosing class
from typing import Tuple, List, Dict
from itertools import chain
import numpy as np
import re
import ipaddress
from pybatfish.datamodel import *


def convert_bgp_route(rec: np.record) -> BgpRoute:
    return BgpRoute(network=rec.Network,
        protocol=rec.Protocol,
        asPath=[[int(hop)] for hop in rec.AS_Path.split(' ')] if rec.AS_Path != "" else [],
        communities=rec.Communities,
        localPreference=rec.Local_Pref,
        metric=rec.Metric,
        originatorIp=rec.Originator_Id,
        originType=rec.Origin_Type,
        sourceProtocol=rec.Origin_Protocol)

def unpack_bgp_route(route: BgpRoute) -> Tuple:
    return (route.network, *chain.from_iterable(route.asPath))

def unpack_as_path(aspath: List[List[int]]) -> Tuple:
    return (*chain.from_iterable(aspath),)

def convert_external_ra(d: Dict) -> BgpRoute:
    return BgpRoute(
        network=d['network'],
        protocol='bgp',
        asPath=d['asPath'],
        communities=d['communities'],
        localPreference=d['localPreference'],
        metric=d["med"],
        originatorIp=d['originatorIp'],
        originType=d['originType'],
        sourceProtocol=d['srcProtocol'],
    )


def convert_list_wrapper(lst: ListWrapper):
    return [a for a in lst]

def convert_prefix(prefix: str) -> ipaddress.ip_network:
    return ipaddress.ip_network(prefix, strict=False)

def convert_ipv4_prefix(prefix: str) -> Tuple[bool, str]:
    ipn = ipaddress.ip_network(prefix, strict=False)
    return (ipn.version == 4), str(ipn.network_address)

def extract_digits(name: str) -> List[int]:
    return [int(s) for s in re.findall(r'\d+', name)]

def get_prefix_len(rule) -> int:
    return rule.prefix.prefixlen

def is_isp(device_name: str) -> bool:
    if device_name == None:
        return False
    return device_name.startswith('isp_')

# example:
# "Matched policy-statement SEND-DEFAULT term REJECT" -> ("policy-statement term", "SEND-DEFAULT REJECT")
def convert_trace_element(trace: TraceElement) -> Tuple(str, str):
    if len(trace.fragments) == 2 and str(trace.fragments[0]) == "Matched ":
        meta = trace.fragments[1]
        words = str(meta).split(' ')
        if len(words) == 4:
            if words[2] == "sequence-number":
                words[2] = "entry"
            return (' '.join([words[0], words[2]]), ' '.join([words[1], words[3]]))
    logging.getLogger(__name__).warning(f"Cannot convert traceElement: {trace}")
    return ("Unknown type", "Unknown clause")

# example:
# "Matched policy-statement SEND-DEFAULT term REJECT" -> "SEND-DEFAULT"
def get_policy_name(trace: TraceElement) -> str:
    if len(trace.fragments) == 2 and str(trace.fragments[0]) == "Matched ":
        meta = trace.fragments[1]
        words = str(meta).split(' ')
        if len(words) == 4:
            return words[1]
    logging.getLogger(__name__).warning(f"Cannot convert traceElement: {trace}")
    return "Unknown name"

# example
# '65401 577 65537' -> ['65401', '577', '65537']
def convert_as_path(as_path: str) -> List[List[int]]:
    if as_path == "" or as_path == None:
        return []
    else:
        return [[int(hop)] for hop in as_path.split(' ')]


def ip_is_in_range(ip: str, range: str) -> bool:
    ipa = ipaddress.ip_network(ip)
    ipn = ipaddress.ip_network(range)
    return ipa.subnet_of(ipn)

def find_matched_range(ip: str, ranges: list) -> str:
    for range in ranges:
        if ip_is_in_range(ip, range):
            return range
    return None

class _CaptureEq:
    'Object wrapper that remembers "other" for successful equality tests.'
    def __init__(self, obj):
        self.obj = obj
        self.match = obj
    def __eq__(self, other):
        result = (self.obj == other)
        if result:
            self.match = other
        return result
    def __hash__(self) -> int:
        return hash(self.obj)
    def __getattr__(self, name):  # support hash() or anything else needed by __contains__
        return getattr(self.obj, name)

def get_equivalent(container, item, default=None):
    '''Gets the specific container element matched by: "item in container".

    Useful for retreiving a canonical value equivalent to "item".  For example, a
    caching or interning application may require fetching a single representative
    instance from many possible equivalent instances).

    >>> get_equivalent(set([1, 2, 3]), 2.0)             # 2.0 is equivalent to 2
    2
    >>> get_equivalent([1, 2, 3], 4, default=0)
    0
    '''
    t = _CaptureEq(item)
    if t in container:
        return t.match
    return default




#   Copyright 2022 Xieyang Xu
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from setuptools import setup, find_packages

setup(
    name='netcov',
    version='0.1.1',
    description='Coverage analyzer for network router configurations',
    author='Xieyang Xu',
    author_email='ashlippers@gmail.com',
    packages=find_packages(exclude=('tests')),
    install_requires=['pandas', 'pickle-mixin', 'networkx', 'pyvis', 'treelib', 'dd', 'parse', 'numpy', 'ipaddress', 'pybatfish'],
)


#   Copyright 2022 Xieyang Xu
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from __future__ import annotations
import ipaddress
import logging
from typing import DefaultDict, Generator, Iterable, Dict, List, Optional
from pybatfish.client.session import Session

from .batchmanager import BatchManager
from .ifg import IFG
from .sortedrib import *
from .sourcelines import *
from .netstate import *
from .utils import *

class Vrf:
    def __init__(self, host: str, vrf: str) -> None:
        self.host: str= host
        self.vrf: str= vrf
        self.name: str = f"{host}.{vrf}"
        self.ribs: Dict[str, IndexedRib] = {}
        self.interfaces: Dict[str, InterfaceConfig] = {}
        self.bgp_enabled: bool = False
        self.bgp_sessions: List[BgpSessionStatus] = []
        self.bgp_peer_configs: List[BgpPeerConfig] = []
        self.bgp_group_configs: Dict[str, BgpGroupConfig] = {}
        self.rib: SortedRib = SortedRib()
        self.source_lines: SourceLines = SourceLines()

    def add_rib(self, key: str, rib: IndexedRib) -> None:
        if key in self.ribs:
            logging.getLogger(__name__).warning(f"WARNING: Duplicated Rib {key} at {self.name}")
        else:
            self.ribs[key] = rib
    
    def get_rib(self, type: str) -> IndexedRib:
        if type in self.ribs:
            return self.ribs[type]
        else:
            logging.getLogger(__name__).warning(f"WARNING: Missing {type} Rib at {self.name}")
            return None
    
    # def get_bgp_announcement_rib(self, direction: str, peer:str) -> Optional[BgpEdge]:
    #     if direction == "export":
    #         for rib in self.ribs.values():
    #             if isinstance(rib, BgpEdge) and rib.sender == self.name and rib.receiver == peer:
    #                 return rib
    #     elif direction == "import":
    #         for rib in self.ribs.values():
    #             if isinstance(rib, BgpEdge) and rib.receiver == self.name and rib.sender == peer:
    #                 return rib
    #     return None

    def find_bgp_session_for_peer(self, peer: str) -> Optional[BgpSessionStatus]:
        for session in self.bgp_sessions:
            if session.peer == peer:
                return session
        return None

    def find_bgp_session_with_as_ip(self, asn: int, ip: str) -> Optional[BgpSessionStatus]:
        for session in self.bgp_sessions:
            if int(session.remote_as) == int(asn) and session.remote_ip == ip:
                return session
        return None

    def find_bgp_peer_config_for_session(self, session: BgpSessionStatus) -> Optional[BgpPeerConfig]:
        for config in self.bgp_peer_configs:
            if isinstance(config, BgpPeerConfigP2p) and config.remote_as == session.remote_as and config.remote_ip == session.remote_ip:
                return config
            elif isinstance(config, BgpPeerConfigPassive) and ip_is_in_range(session.remote_ip, config.listen_range):
                return config
        return None

    def resolve_interface_for_ip(self, ip: str) -> Iterable[InterfaceConfig]:
        try:
            ipa = ipaddress.ip_address(ip)
        except:
            return []
        
        interfaces = self.rib.forward(ip)
        return [self.interfaces[i] for i in interfaces]

    def resolve_rib_rules_for_ip(self, ip:str) -> Iterable[RibRule]:
        try:
            ipa = ipaddress.ip_address(ip)
        except:
            return []
        
        rules = self.rib.matched_rules(ip)
        return rules

class Device:
    def __init__(self, name: str):
        self.name: str = name
        self.vrfs: Dict[str, Vrf] = {}
        self.is_isp: bool = is_isp(name)
        self.routemaps: Dict[str, Routemap] = {}
        self.raw_policies: List[str] = []
        self.interface_configs: Dict[str, InterfaceConfig] = {}
        self.referenced_configs: Dict[Tuple[str, str], ReferencedConfig] = {}
        self.referenced_lines: Dict[int, ReferencedConfig] = {}
        self.raw_bgp_groups: Dict[str, BgpGroupConfigRaw] = {}
        self.raw_routemap_clauses: DefaultDict[str, List[RoutemapClause]] = defaultdict(list)

    def add_vrf(self, name: str, vrf: Vrf) -> None:
        if name in self.vrfs:
            logging.getLogger(__name__).warning(f"WARNING: Duplicated VRF {name} at {self.name}")
        else:
            self.vrfs[name] = vrf

    def get_vrf(self, vrf: str = "default") -> Optional[Vrf]:
        return self.vrfs.get(vrf, None)

    def get_routemap(self, routemap: str) -> Optional[Routemap]:
        return self.routemaps.get(routemap, None)

    def get_interface_config(self, name: str) -> Optional[InterfaceConfig]:
        return self.interface_configs.get(name, None)

    def find_bgp_session_with_as_ip(self, remote_as: int, remote_ip: str) -> Optional[BgpSessionStatus]:
        for vrf in self.vrfs.values():
            for session in vrf.bgp_sessions:
                if int(session.remote_as) == int(remote_as) and session.remote_ip == remote_ip:
                    return session
        return None

    def find_bgp_peer_with_ip(self, remote_ip: str, vrf_name: str) -> Optional[BgpPeerConfig]:
        vrf = self.get_vrf(vrf_name)
        if vrf is not None:
            for session in vrf.bgp_sessions:
                if session.remote_ip == remote_ip:
                    return vrf.find_bgp_peer_config_for_session(session)
        return None
      
class Network:
    def __init__(self, session: Session, snapshot_path: str, static_analysis: bool = False):
        self.bf: Session = session
        self.inited: bool = False
        self.snapshot_path: str = snapshot_path
        self.static_analysis: bool = static_analysis
        self.devices: Dict[str, Device] = {}
        self.filenames: Dict[str, str] = {}
        self.devicenames: Dict[str, str] = {}
        self.bgp_edges: Dict[Tuple[str, str, str, str], BgpEdge] = {}
        self.graph: IFG = IFG()
        self.bm: BatchManager = BatchManager(session)
        self.source: SourceLines = SourceLines()
        self.supported_source: SourceLines = SourceLines()
        self.dead_source: SourceLines = SourceLines()
        self.reachable_source: SourceLines = SourceLines()
        self.typed_source = defaultdict(SourceLines)
        self.cnt_routemaps: int = -1
        self.cnt_routemap_clauses: int = -1
        self.cnt_bgp_peer_configs: int = -1
        self.cnt_rib_entry: int = -1
        self.cnt_interface: int = -1


    def iter_vrfs(self) -> Generator[Tuple[str, str, Device, Vrf]]:
        for device_name, device in self.devices.items():
            for vrf_name, vrf in device.vrfs.items():
                yield device_name, vrf_name, device, vrf

    def get_vrf_by_name(self, name: str) -> Optional[Vrf]:
        words = name.split(".")
        if len(words) == 2:
            device = self.devices.get(words[0], None)
            if device != None:
                return device.get_vrf(words[1], None)
        return None

    def get_vrf(self, host: str, vrf: str) -> Optional[Vrf]:
        device = self.devices.get(host, None)
        if device != None:
            return device.get_vrf(vrf)
        return None

    def devicename_to_filename(self, device_name: str) -> Optional[str]:
        return self.filenames.get(device_name, None)

    def filename_to_device_name(self, filename: str) -> Optional[str]:
        return self.devicenames.get(filename, None)

    def add_bgp_edge(self, edge: BgpEdge) -> None:
        tup = (edge.sender, edge.sender_vrf, edge.receiver, edge.receiver_vrf)
        if tup not in self.bgp_edges:
            self.bgp_edges[tup] = edge
        else:
            logging.getLogger(__name__).warning(f"WARNING: Duplicated BGP edge {tup}")

    def get_bgp_edge(self, s_host: str, s_vrf: str, r_host: str, r_vrf: str) -> Optional[BgpEdge]:
        tup = (s_host, s_vrf, r_host, r_vrf)
        return self.bgp_edges.get(tup, None)




#   Copyright 2022 Xieyang Xu
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from __future__ import annotations # type hint in the enclosing class
import json
import logging
from typing import Iterable, Optional, Set
from pybatfish.datamodel.acl import TraceTreeList, TraceElement
import numpy as np

from .netstate import *
from .utils import *

class DNode:
    def __init__(self) -> None:
        self.pred: Set[Tuple[DNode, bool]] = set()
        self.status: Set[str] = {"init"}

    def add_dependency(self, node: DNode, is_weak=False):
        self.pred.add((node, is_weak))

    def print_dependency(self):
        res = [f"{pred} {'weak' if weak else 'strong'}" for (pred, weak) in self.pred]
        return "\n\t".join(res)
    
    def __eq__(self, __o: object) -> bool:
        if isinstance(__o, DNode):
            return self.__hash__() == __o.__hash__()
        return NotImplemented

class DataplaneTestNode(DNode):
    def __init__(self, test_type, host, dst_ip, prefix, nexthop_ip, nexthop_interface) -> None:
        super().__init__()
        self.test_type: str = test_type
        self.host: str = host
        self.dst_ip: str = dst_ip
        self.prefix: str = prefix
        self.nexthop_ip: str = nexthop_ip
        self.nexthop_interface: str = nexthop_interface

    def __repr__(self) -> str:
        if self.test_type == "forwarded_to_ip":
            content = f"{self.dst_ip} -> {self.nexthop_interface}({self.nexthop_ip})"
        elif self.test_type == "forwarded_to_interface":
            content = f"{self.dst_ip} -> {self.nexthop_interface}"
        else:
            content = f"Unknown test type: {self.test_type}"
        
        return f"Tested@{self.host} {content}"

    def __hash__(self) -> int:
        return hash((self.__class__.__name__, self.test_type, self.host, self.dst_ip, self.prefix, self.nexthop_ip, self.nexthop_interface))
    
    def toJson(self):
        d = {
            "test_type": self.test_type,
            "host": self.host,
            "dst_ip": self.dst_ip,
            "prefix": self.prefix,
            "nexthop_ip": self.nexthop_ip,
            "nexthop_interface": self.nexthop_interface,
        }
        return json.dumps(d)

    def from_dict(d: dict) -> DataplaneTestNode:
        return DataplaneTestNode(d["test_type"], d["host"], d["dst_ip"], d["prefix"], d["nexthop_ip"], d["nexthop_interface"])

class RouteNode(DNode):
    """Base class of route facts
    """
    def __init__(self, rib_type, prefix, nexthop, is_drop) -> None:
        super().__init__()
        self.rib_type: str = rib_type
        self.prefix: str = prefix
        self.nexthop: Optional[str] = nexthop
        self.is_drop: bool = is_drop

class MainRouteNode(RouteNode):
    def __init__(self, rib_type, prefix, nexthop, host, vrf, protocol, is_drop = False) -> None:
        super().__init__(rib_type, prefix, nexthop, is_drop)
        self.host: str = host
        self.vrf: str = vrf
        self.protocol: str = protocol

    def from_rec(rec: np.record) -> Optional[MainRouteNode]:
        if rec.Protocol in ['connected', 'local', 'static']:
            node_main = MainRouteNode("main", rec.Network, rec.Next_Hop_Interface, rec.Node, rec.VRF, rec.Protocol)
        elif rec.Protocol in ['bgp', 'ibgp', 'aggregate', 'isisL1', 'isisL2', 'isisEL1']:
            nexthop = rec.Next_Hop_IP
            if nexthop == None or nexthop == "AUTO/NONE(-1l)":
                nexthop = rec.Next_Hop_Interface
            node_main = MainRouteNode("main", rec.Network, nexthop, rec.Node, rec.VRF, rec.Protocol)
        else:
            logging.getLogger(__name__).warning(f"Unsupported protocol {rec.Protocol} with prefix {rec.Network} at {rec.Node}.{rec.VRF} main rib")
            node_main = None
        return node_main
    
    def __repr__(self) -> str:
        drop = "[drop]" if self.is_drop else ""
        return f"MainRoute@{self.host}.{self.vrf} {self.protocol}: {self.prefix} -> {self.nexthop}" + drop

    def __hash__(self) -> int:
        return hash((self.__class__.__name__, self.rib_type, self.prefix, self.nexthop, self.host, self.vrf, self.protocol, self.is_drop))

class BgpAnnouncementNode(RouteNode):
    def __init__(self, rib_type, prefix, nexthop, sender, receiver, sender_vrf, receiver_vrf, pred_route, bgp_edge, trace, succ_node, is_drop = False) -> None:
        super().__init__(rib_type, prefix, nexthop, is_drop)
        self.sender: str = sender
        self.receiver: str = receiver
        self.sender_vrf: str = sender_vrf
        self.receiver_vrf: str = receiver_vrf
        # atributes for bgp export evaluation
        self.pred_route: BgpRoute = pred_route
        self.bgp_edge: Optional[BgpEdge] = bgp_edge
        self.trace: Optional[TraceTreeList] = trace
        # atribute to trigger bgp import evaluation at succ node
        # after export evaluation is done
        self.succ_node: Optional[BgpRouteNode] = succ_node

    def __repr__(self) -> str:
        drop = "[drop]" if self.is_drop else ""
        return f"RA@{self.sender}.{self.sender_vrf}->{self.receiver}.{self.receiver_vrf} {self.prefix}" + drop

    def __hash__(self) -> int:
        #return hash((self.__class__.__name__, self.rib_type, self.prefix, self.nexthop, self.sender, self.receiver,\
        #    self.sender_vrf, self.receiver_vrf, unpack_bgp_route(self.pred_route), self.is_drop))
        return hash((self.__class__.__name__, self.rib_type, self.prefix, self.nexthop, self.sender, self.receiver,\
            self.sender_vrf, self.receiver_vrf, self.is_drop))

class BgpRouteNode(RouteNode):
    def __init__(self, rib_type, prefix, nexthop, host, vrf, protocol, origin_protocol, as_path, route_status, received_from_ip, is_drop = False) -> None:
        super().__init__(rib_type, prefix, nexthop, is_drop)
        self.host: str = host
        self.vrf: str = vrf
        self.protocol: str = protocol
        self.origin_protocol: str = origin_protocol
        self.as_path: List[List[int]] = as_path
        self.route_status: str = route_status
        self.received_from_ip: str = received_from_ip
        # atributes for bgp import evaluation
        self.pred_route: Optional[BgpRoute] = None
        self.import_policy: Optional[Iterable[str]] = None
        self.trace: Optional[TraceTreeList] = None
        # atribute for "bgp_from_border_bgp_session"
        self.from_session: Optional[BgpSessionStatus] = None

    def from_rec(rec: np.record) -> Optional[BgpRouteNode]:
        if rec.Protocol == "bgp" or rec.Protocol == "aggregate" or rec.Protocol == "ibgp":
            if rec.Origin_Protocol == "connected":
                nexthop = rec.Next_Hop_Interface
                if nexthop == "null_interface":
                    nexthop = rec.Next_Hop_IP
            else:
                nexthop = rec.Next_Hop_IP
                if nexthop == None:
                    nexthop = rec.Next_Hop_Interface
            node_bgp = BgpRouteNode("bgp", rec.Network, nexthop, rec.Node, rec.VRF, rec.Protocol, rec.Origin_Protocol, convert_as_path(rec.AS_Path), rec.Status[0], rec.Received_From_IP)
        else:
            logging.getLogger(__name__).warning(f"Unsupported protocol {rec.Protocol} with prefix {rec.Network} at {rec.Node}.{rec.VRF} bgp")
            node_bgp = None
        return node_bgp

    def __repr__(self) -> str:
        drop = "[drop]" if self.is_drop else ""
        return f"BGPRoute@{self.host}.{self.vrf} {self.prefix} -> {self.nexthop}" + drop

    def __hash__(self) -> int:
        return hash((self.__class__.__name__, self.rib_type, self.prefix, self.nexthop, self.host, self.vrf, self.protocol,\
            self.origin_protocol, unpack_as_path(self.as_path), self.route_status, self.is_drop))
    
    def top_of_as_path_matches(self, nexthop_as: int) -> bool:
        if len(self.as_path) > 0 and len(self.as_path[0]) > 0 and self.as_path[0][0] == nexthop_as:
            return True
        return False

class ConnectedRouteNode(RouteNode):
    def __init__(self, rib_type, prefix, nexthop, host, vrf, protocol, is_drop = False) -> None:
        super().__init__(rib_type, prefix, nexthop, is_drop)
        self.host: str = host
        self.vrf: str = vrf
        self.protocol: str = protocol

    def from_rec(rec: np.record) -> Optional[ConnectedRouteNode]:
        if rec.Protocol == 'connected' or rec.Protocol == "local" or rec.Protocol == "static":
            node_connected = ConnectedRouteNode("connected", rec.Network, rec.Next_Hop_Interface, rec.Node, rec.VRF, rec.Protocol)
        else:
            logging.getLogger(__name__).warning(f"Unsupported protocol {rec.Protocol} with prefix {rec.Prefix} at {rec.Node}.{rec.VRF} connected rib")
            node_connected = None
        return node_connected

    def __repr__(self) -> str:
        drop = "[drop]" if self.is_drop else ""
        return f"ConnectedRoute@{self.host}.{self.vrf}({self.protocol}) {self.prefix} -> {self.nexthop}" + drop

    def __hash__(self) -> int:
        return hash((self.__class__.__name__, self.rib_type, self.prefix, self.nexthop, self.host, self.vrf, self.protocol, self.is_drop))

class ConfigNode(DNode):
    """base class of config facts"""
    def __init__(self, host, config_type) -> None:
        super().__init__()
        self.host: str = host
        self.config_type: str = config_type

class UserSuppliedConfigNode(ConfigNode):
    def __init__(self, host, lines, config_type='user-supplied') -> None:
        super().__init__(host, config_type)
        self.lines: FileLines = lines

    def __repr__(self) -> str:
        return f"{self.config_type}@{self.host}"

    def __hash__(self) -> int:
        return hash((self.__class__.__name__, self.host, self.config_type, *self.lines))

class RoutemapNode(ConfigNode):
    def __init__(self, host, config_type, name, lines) -> None:
        super().__init__(host, config_type)
        self.name: str = name
        self.lines: FileLines = lines

    def __repr__(self) -> str:
        return f"Routemap@{self.host} {self.name}"

    def __hash__(self) -> int:
        return hash((self.__class__.__name__, self.host, self.config_type, self.name))

class RoutemapClauseNode(ConfigNode):
    def __init__(self, host, config_type, name, seq, lines) -> None:
        super().__init__(host, config_type)
        self.name: str = name
        self.seq: str = seq
        self.lines: FileLines = lines

    def __repr__(self) -> str:
        return f"RM-Clause@{self.host} {self.name}"

    def __hash__(self) -> int:
        return hash((self.__class__.__name__, self.host, self.config_type, self.name, self.seq))

class BgpPeerConfigNode(ConfigNode):
    def __init__(self, host, vrf, config_type, is_passive, local_as, export_policies, import_policies, peer_group, lines) -> None:
        super().__init__(host, config_type)
        self.vrf: str = vrf
        self.is_passive: bool = is_passive
        self.local_as: int = local_as
        self.export_policies: List[str] = export_policies
        self.import_policies: List[str] = import_policies
        self.peer_group: str = peer_group
        self.lines: FileLines = lines

class BgpPeerConfigP2pNode(BgpPeerConfigNode):
    def __init__(self, host, vrf, config_type, is_passive, local_as, export_policies, import_policies, peer_group, lines, local_ip, remote_ip, remote_as) -> None:
        super().__init__(host, vrf, config_type, is_passive, local_as, export_policies, import_policies, peer_group, lines)
        self.local_ip: str = local_ip
        self.remote_ip: str = remote_ip
        self.remote_as: int = remote_as
    
    def from_config(config: BgpPeerConfigP2p):
        return BgpPeerConfigP2pNode(config.host, config.vrf, "bgp_peer_config_p2p", config.is_passive, config.local_as, config.export_policies, config.import_policies, config.peer_group, config.lines, config.local_ip, config.remote_ip, config.remote_as)
    
    def __repr__(self) -> str:
        return f"BgpPeerP2pConfig@{self.host}.{self.vrf} peer:{self.remote_as}@{self.remote_ip}"

    def __hash__(self) -> int:
        return hash((self.__class__.__name__, self.host, self.vrf, self.config_type, self.is_passive, self.local_as, self.local_ip, self.remote_ip, self.remote_as, self.peer_group))

class BgpPeerConfigPassiveNode(BgpPeerConfigNode):
    def __init__(self, host, vrf, config_type, is_passive, local_as, export_policies, import_policies, peer_group, lines, listen_range) -> None:
        super().__init__(host, vrf, config_type, is_passive, local_as, export_policies, import_policies, peer_group, lines)
        self.listen_range: str = listen_range

    def from_config(config: BgpPeerConfigPassive):
        return BgpPeerConfigPassiveNode(config.host, config.vrf, "bgp_peer_config_passive", config.is_passive, config.local_as, config.export_policies, config.import_policies, config.peer_group, config.lines, config.listen_range)

    def __repr__(self) -> str:
        return f"BgpPeerPassiveConfig@{self.host}.{self.vrf} listen range:{self.listen_range}"

    def __hash__(self) -> int:
        return hash((self.__class__.__name__, self.host, self.vrf, self.config_type, self.is_passive, self.local_as, self.export_policies, self.import_policies, self.peer_group, self.listen_range))

class InterfaceConfigNode(ConfigNode):
    def __init__(self, host, config_type, name, lines) -> None:
        super().__init__(host, config_type)
        self.name: str = name
        self.lines: FileLines = lines

    def from_interface(it: InterfaceConfig) -> InterfaceConfigNode:
        return InterfaceConfigNode(it.host, "interface", it.name, it.lines)

    def __repr__(self) -> str:
        return f"Interface[{self.name}]@{self.host}" 

    def __hash__(self) -> int:
        return hash((self.__class__.__name__, self.host, self.config_type, self.name))

class ReferencedConfigNode(ConfigNode):
    def __init__(self, host, config_type, name, lines) -> None:
        super().__init__(host, config_type)
        self.name: str = name
        self.lines: FileLines = lines

    def from_config(config: ReferencedConfig) -> ReferencedConfigNode:
        return ReferencedConfigNode(config.host, config.config_type, config.name, config.lines)

    def __repr__(self) -> str:
        return f"{self.config_type.title()}@{self.host} {self.name}" 

    def __hash__(self) -> int:
        return hash((self.__class__.__name__, self.host, self.config_type, self.name))

# class MatchRoutemapNode(DNode):
#     def __init__(self, route, routemap, trace) -> None:
#         super().__init__()
#         self.route: BgpRoute = route
#         self.routemap: Routemap = routemap
#         self.trace: TraceTreeList = trace

#     def __repr__(self) -> str:
#         return f"Match({self.routemap}, BgpRoute(network={self.route.network}, asPath={self.route.asPath}))"

#     def __hash__(self) -> int:
#         return hash((self.__class__.__name__, unpack_bgp_route(self.route), self.routemap.host, self.routemap.name))

# class MatchRoutemapClauseNode(DNode):
#     def __init__(self, route, clause) -> None:
#         super().__init__()
#         self.route: BgpRoute = route
#         self.clause: RoutemapClause = clause

#     def __repr__(self) -> str:
#         return f"Match({self.clause}, BgpRoute(network={self.route.network}, asPath={self.route.asPath}))"

#     def __hash__(self) -> int:
#         return hash((self.__class__.__name__, unpack_bgp_route(self.route), self.clause.host, self.clause.name))

# class MatchKeywordNode(DNode):
#     def __init__(self, route, referenced_config) -> None:
#         super().__init__()
#         self.route: BgpRoute = route
#         self.referenced_config: ReferencedConfig = referenced_config

#     def __repr__(self) -> str:
#         return f"Match({self.referenced_config}, BgpRoute(network={self.route.network}, asPath={self.route.asPath}))"

#     def __hash__(self) -> int:
#         return hash((self.__class__.__name__, unpack_bgp_route(self.route), self.referenced_config.host, self.referenced_config.config_type, self.referenced_config.name))

class EstablishedBgpSessionNode(DNode):
    def __init__(self, host, vrf, peer, peer_vrf, local_as, remote_as, local_ip, remote_ip, local_interface, remote_interface, session_type, is_border) -> None:
        super().__init__()
        self.host: str = host
        self.vrf: str = vrf
        self.peer: str = peer
        self.peer_vrf: str = peer_vrf
        self.local_as: int = local_as
        self.remote_as: int = remote_as
        self.local_ip: Optional[str] = local_ip
        self.remote_ip: Optional[str] = remote_ip
        self.local_interface: Optional[str] = local_interface
        self.remote_interface: Optional[str] = remote_interface
        self.session_type: str = session_type
        self.is_border: bool = is_border

    def from_session(s: BgpSessionStatus, peer_vrf: str) -> Optional[EstablishedBgpSessionNode]:
        if s.established_status == "ESTABLISHED":
            return EstablishedBgpSessionNode(s.host, s.vrf, s.peer, peer_vrf, s.local_as, s.remote_as, s.local_ip, s.remote_ip, s.local_interface, s.remote_interface, s.session_type, s.is_border)
        else:
            return None

    def __repr__(self) -> str:
        return f"BgpSession@{self.host}.{self.vrf}({self.local_as}, {self.local_ip})-{self.peer}({self.remote_as}, {self.remote_ip})"

    def __hash__(self) -> int:
        return hash((self.__class__.__name__, self.host, self.vrf, self.peer, self.local_as, self.remote_as, self.local_ip, self.remote_ip, self.local_interface, self.remote_interface, self.session_type))

# class L3BiConnectivityNode(DNode):
#     def __init__(self, host1, host2, vrf1, vrf2, ip1, ip2, ipprotocol, port1, port2) -> None:
#         super().__init__()
#         self.host1: str = host1
#         self.host2: str = host2
#         self.vrf1: str = vrf1
#         self.vrf2: str = vrf2
#         self.ip1: str = ip1
#         self.ip2: str = ip2
#         self.ipprocotol: str = ipprotocol
#         self.port1: str = port1
#         self.port2: str = port2

#     def __repr__(self) -> str:
#         return f"L3Connectivity: ({self.host1}.{self.vrf1}, {self.ip1}) <--> ({self.host2}.{self.vrf2}, {self.ip2})"

#     def forward(self) -> str:
#         return f"L3Connectivity: ({self.host1}.{self.vrf1}, {self.ip1}) -> ({self.host2}.{self.vrf2}, {self.ip2})"

#     def backward(self) -> str:
#         return f"L3Connectivity: ({self.host2}.{self.vrf2}, {self.ip2}) -> ({self.host1}.{self.vrf1}, {self.ip1})"

#     def __hash__(self) -> int:
#         return hash((self.__class__.__name__, frozenset([(self.host1, self.vrf1, self.ip1, self.port1), (self.host2, self.vrf2, self.ip2, self.port2)]), self.ipprocotol))

class L3ConnectivityNode(DNode):
    def __init__(self, host1, host2, vrf1, vrf2, ip1, ip2, ipprotocol, port1, port2) -> None:
        super().__init__()
        self.host1: str = host1
        self.host2: str = host2
        self.vrf1: str = vrf1
        self.vrf2: str = vrf2
        self.ip1: str = ip1
        self.ip2: str = ip2
        self.ipprocotol: str = ipprotocol
        self.port1: str = port1
        self.port2: str = port2

    def __repr__(self) -> str:
        return f"L3Connectivity: ({self.host1}.{self.vrf1}, {self.ip1}) -> ({self.host2}.{self.vrf2}, {self.ip2})"

    def __hash__(self) -> int:
        return hash((self.__class__.__name__, self.host1, self.vrf1, self.ip1, self.port1, self.host2, self.vrf2, self.ip2, self.port2, self.ipprocotol))

class L3PathNode(DNode):
    def __init__(self, dst_ip, path) -> None:
        super().__init__()
        self.dst_ip: str = dst_ip
        self.path: Trace = path
        hops = []
        for hop in self.path:
            host = hop.node
            for step in hop.steps:
                if step.action == "FORWARDED":
                    hops.append(f"{host}[{step.detail.forwardingDetail.outputInterface}]")
                elif step.action == "ACCEPTED":
                    hops.append(f"{host}[{step.detail.interface}]")
        self.repr: str = ' -> '.join(hops)

    def __repr__(self) -> str:
        return f"L3Path: {self.dst_ip}@{self.repr}"

    def __hash__(self) -> int:
        return hash((self.__class__.__name__, self.dst_ip, self.repr))

    